#!/usr/bin/perl
use strict;
use warnings;
use lib 'lib';

use FileHandle ();
use DirHandle ();
use File::Copy qw(copy);
use File::Path qw(mkpath);
use JSON::PP ();
use Getopt::Long qw(GetOptions);

my $json = JSON::PP->new();
$json->canonical(1);
$json->pretty(1);
$json->relaxed(1);

my($no_ask,$textui);
GetOptions(
	"no-ask" => \$no_ask,
	"text" => \$textui,
);

my $have_tk;
eval {
	require Tkx;
	$have_tk = 1;
} unless $no_ask or $textui;

my $initconf = shift @ARGV || 'bluestem.conf';
my $fh = FileHandle->new($initconf)
	or die "Could not open config file `bluestem.conf': $!";
my $conf = $json->decode(join('',<$fh>));

my $install_dir = $conf->{install_dir};
my $cgi_install_dir = $conf->{cgi_install_dir} || '';
my $cgipath = $conf->{cgipath};
my $img_install_dir = $conf->{img_install_dir} || '';
my $imgpath = $conf->{imgpath};
my $asp_install_dir = $conf->{asp_install_dir} || '';
my $asppath = $conf->{asppath};
my $servername = '';
my $server_key = '';

sub rec_copy ($$);
sub rec_copy ($$)
{
	my($src,$dest) = @_;
	-d $dest or mkpath $dest;
	my $dh = DirHandle->new($src);
	for my $file ( $dh->read() ) {
		next unless $file =~ m/[a-zA-Z]/;
		if( -d "$src/$file" ) {
			rec_copy "$src/$file", "$dest/$file";
		} else {
			print "Copying $src/$file to $dest/$file\n";
			copy "$src/$file", "$dest/$file";
		}
	}
}

sub ask ($;$)
{
	my($q,$d) = @_;
	return $d if $no_ask;

	if( $d ) {
		print "$q [$d] ";
	} else {
		print "$q ";
	}
	my $a = <STDIN>;
	chomp $a;
	return $a || $d;
}

sub do_install ()
{
	my($fh,$dh,$content);
	{	local $/; # Slurp mode
		$servername or die "No servername specified!\n";
		$server_key or die "No server_key specified!\n";
		$conf->{install_dir} = $install_dir;
		$conf->{cgi_install_dir} = $cgi_install_dir;
		$conf->{cgipath} = $cgipath;
		$conf->{img_install_dir} = $img_install_dir;
		$conf->{imgpath} = $imgpath;
		$conf->{servername} = $servername;
		$conf->{server_key} = $server_key;
		$conf->{asp_install_dir} = $asp_install_dir;
		$conf->{asppath} = $asppath;

		for my $k ( keys %$conf ) {
			$conf->{$k} =~ s/INSTALL_DIR/$install_dir/;
			$conf->{$k} =~ s{//+}{/}g;
		}

		for my $dir (
			$conf->{install_dir},     $conf->{asp_install_dir},
			$conf->{cgi_install_dir}, $conf->{img_install_dir},
			$conf->{cache_dir},       $conf->{asp_cache_dir},
		) {
			-d $dir or mkpath $dir;
		}

		my $conf_file = "$install_dir/bluestem.conf";
		$fh = FileHandle->new($conf_file,">")
			or die "Failed to open $conf_file";
		$fh->print( $json->encode($conf),"\n" );

		rec_copy("template",$conf->{template_dir});
		rec_copy("img",$conf->{img_install_dir});
		rec_copy("lib","$install_dir/lib");

		# Copy over all the cgi scripts, making sure to change the
		# "use lib" statements
		$dh = DirHandle->new("cgi");
		for my $cgi ( $dh->read() ) {
			next if $cgi =~ m/^\./;
			$fh = FileHandle->new("cgi/$cgi");
			$content = <$fh>;
			$content =~ s{/usr/local/bluestem-client/lib}{$install_dir/lib};
			$fh = FileHandle->new("$conf->{cgi_install_dir}/$cgi",">")
				or die "Could not write $conf->{cgi_install_dir}/$cgi: $!";
			$fh->print($content);
			$fh->close();
		}

		# Copy over all the asp scripts, doing string substitution
		$dh = DirHandle->new("asp");
		for my $asp ( $dh->read() ) {
			next if $asp =~ m/^\./;
			$fh = FileHandle->new("asp/$asp");
			$content = <$fh>;
			$content =~ s/\%ASPVirtDir\%/$conf->{asppath}/g;
			$content =~ s/\%BluestemCgiDir\%/$conf->{cgipath}/g;
			$content =~ s/\%BluestemCache\%/$conf->{cache_dir}/g;
			$content =~ s/\%BluestemAspCache\%/$conf->{asp_cache_dir}/g;
			$fh = FileHandle->new("$conf->{asp_install_dir}/$asp",">")
				or die "Could not write $conf->{asp_install_dir}/$asp: $!";
			$fh->print($content);
			$fh->close();
		}

		$fh = FileHandle->new("lib/Bluestem/Client.pm");
		$content = <$fh>;
		$content =~ s{/usr/local/bluestem-client/bluestem.conf}{$conf_file};
		$fh = FileHandle->new("$install_dir/lib/Bluestem/Client.pm",">");
		$fh->print($content);
		$fh->close();

		$fh = FileHandle->new("bluestem-client-admin.pl");
		$content = <$fh>;
		$content =~ s{/usr/local/bluestem-client/lib}{$install_dir/lib};
		$fh = FileHandle->new("$install_dir/bluestem-client-admin.pl",">");
		$fh->print($content);
		$fh->close();

		Tkx::destroy(".") if $have_tk;
	};

	print "done.\n";
	print "<press enter to exit>\n";
	<STDIN>;
}

if( $have_tk ) {
	$cgi_install_dir ||= "$install_dir/cgi";
	$img_install_dir ||= "$install_dir/img";
	$asp_install_dir ||= "$install_dir/asp";
	Tkx::wm_title(".","Bluestem Client Installation");
	Tkx::ttk__frame(".c", -padding => "3 3 12 12");
	Tkx::grid(".c",-column => 0, -row => 0, -sticky => "nesw");
	Tkx::grid_columnconfigure(".", 0, -weight => 1 );
	Tkx::grid_rowconfigure(".", 0, -weight => 1);
	Tkx::font_create("LabelFont", -family => "Helvetica", -size => 10, -weight => "bold");

	# install_dir
	Tkx::grid( Tkx::ttk__label(".c.install_dir_label", -text => "Target installation folder", -font => 'LabelFont' ), -column => 1, -row => 1, -sticky => "w" );
	Tkx::ttk__entry(".c.install_dir", -width=> 40, -textvariable=>\$install_dir);
	Tkx::grid(".c.install_dir", -column => 2, -row => 1, -sticky => "we");
	Tkx::ttk__button(".c.install_dir_choose",
		-text => "Choose",
		-command => sub {
			$install_dir = Tkx::tk___chooseDirectory(-initialdir=>$install_dir);
			$install_dir =~ s{/$}{};
			$install_dir =~ s{\\}{/}g;
			$asp_install_dir = "$install_dir/asp";
			$cgi_install_dir = "$install_dir/cgi";
			$img_install_dir = "$install_dir/img";
		}
	);
	Tkx::grid(".c.install_dir_choose", -column => 3, -row => 1, -sticky => 'w');
	Tkx::grid( Tkx::ttk__label(".c.install_dir_desc", -text => "The default location for all Bluestem components."), -column => 1, -row => 2, -columnspan => 3, -sticky => "w" );

	# cgi_install_dir
	Tkx::grid( Tkx::ttk__label(".c.cgi_install_dir_label", -text => "CGI installation folder", -font => 'LabelFont'), -column => 1, -row => 3, -sticky => "w" );
	Tkx::ttk__entry(".c.cgi_install_dir", -width=> 40, -textvariable=>\$cgi_install_dir);
	Tkx::grid(".c.cgi_install_dir", -column => 2, -row => 3, -sticky => "we");
	Tkx::ttk__button(".c.cgi_install_dir_choose",
		-text => "Choose",
		-command => sub {$cgi_install_dir = Tkx::tk___chooseDirectory(-initialdir=>$cgi_install_dir);}
	);
	Tkx::grid(".c.cgi_install_dir_choose", -column => 3, -row => 3, -sticky => 'w');
	Tkx::grid( Tkx::ttk__label(".c.cgi_install_dir_desc", -text => "The folder for the Bluestem CGI scripts.  It is recommended that Bluestem CGI scripts be\nkept separate from other CGI scripts on your server."), -column => 1, -row => 4, -columnspan => 3, -sticky => "w" );

	# cgipath
	Tkx::grid( Tkx::ttk__label(".c.cgipath_label", -text => "CGI web path", -font => 'LabelFont'), -column => 1, -row => 5, -sticky => "w" );
	Tkx::ttk__entry(".c.cgipath", -width=> 40, -textvariable=>\$cgipath);
	Tkx::grid(".c.cgipath", -column => 2, -row => 5, -sticky => "we");
	Tkx::grid( Tkx::ttk__label(".c.cgipath_desc", -text => "The path on the webserver to the Bluestem CGI scripts.  A virtual directory should be\nconfigured in IIS from this path to the CGI installation folder."), -column => 1, -row => 6, -columnspan => 3, -sticky => "w" );

	# img_install_dir
	Tkx::grid( Tkx::ttk__label(".c.img_install_dir_label", "-text" => "Image Installation Folder", -font => 'LabelFont'), -column => 1, -row => 7, -sticky => "w" );
	Tkx::ttk__entry(".c.img_install_dir", -width=> 40, -textvariable=>\$img_install_dir);
	Tkx::grid(".c.img_install_dir", -column => 2, -row => 7, -sticky => "we");
	Tkx::ttk__button(".c.img_install_dir_choose",
		-text => "Choose",
		-command => sub {$img_install_dir = Tkx::tk___chooseDirectory(-initialdir=>$img_install_dir);}
	);
	Tkx::grid(".c.img_install_dir_choose", -column => 3, -row => 7, -sticky => 'w');
	Tkx::grid( Tkx::ttk__label(".c.img_install_dir_desc", -text => "The folder for Bluestem images."), -column => 1, -row => 8, -columnspan => 3, -sticky => "w" );

	# imgpath
	Tkx::grid( Tkx::ttk__label(".c.imgpath_label", -text => "Image web path", -font => 'LabelFont'), -column => 1, -row => 9, -sticky => "w" );
	Tkx::ttk__entry(".c.imgpath", -width=> 40, -textvariable=>\$imgpath);
	Tkx::grid(".c.imgpath", -column => 2, -row => 9, -sticky => "we");
	Tkx::grid( Tkx::ttk__label(".c.imgpath_desc", -text => "The path on the webserver to the Bluestem images.  A virtual directory should be\nconfigured in IIS from this path to the image installation folder."), -column => 1, -row => 10, -columnspan => 3, -sticky => "w" );

	# asp_install_dir
	Tkx::grid( Tkx::ttk__label(".c.asp_install_dir_label", "-text" => "ASP Installation Folder", -font => 'LabelFont'), -column => 1, -row => 11, -sticky => "w" );
	Tkx::ttk__entry(".c.asp_install_dir", -width=> 40, -textvariable=>\$asp_install_dir);
	Tkx::grid(".c.asp_install_dir", -column => 2, -row => 11, -sticky => "we");
	Tkx::ttk__button(".c.asp_install_dir_choose",
		-text => "Choose",
		-command => sub {$asp_install_dir = Tkx::tk___chooseDirectory(-initialdir=>$asp_install_dir);}
	);
	Tkx::grid(".c.asp_install_dir_choose", -column => 3, -row => 11, -sticky => 'w');
	Tkx::grid( Tkx::ttk__label(".c.asp_install_dir_desc", -text => "The folder for Classic ASP includes and examples."), -column => 1, -row => 12, -columnspan => 3, -sticky => "w" );


	# asppath
	Tkx::grid( Tkx::ttk__label(".c.asppath_label", -text => "ASP web path", -font => 'LabelFont'), -column => 1, -row => 13, -sticky => "w" );
	Tkx::ttk__entry(".c.asppath", -width=> 40, -textvariable=>\$asppath);
	Tkx::grid(".c.asppath", -column => 2, -row => 13, -sticky => "we");
	Tkx::grid( Tkx::ttk__label(".c.asppath_desc", -text => "The path on the webserver to the Bluestem ASP includes.  A virtual directory should be\nconfigured in IIS from this path to the ASP installation folder."), -column => 1, -row => 14, -columnspan => 3, -sticky => "w" );

	# servername
	Tkx::grid( Tkx::ttk__label(".c.servername_label", -text => "Site Hostname", -font => 'LabelFont'), -column => 1, -row => 15, -sticky => "w" );
	Tkx::ttk__entry(".c.servername", -width=> 40, -textvariable=>\$servername);
	Tkx::grid(".c.servername", -column => 2, -row => 15, -sticky => "we");
	Tkx::grid( Tkx::ttk__label(".c.servername_desc", -text => "This is the hostname used to access the website.  This may or may not be the same as the\nserver hostname."), -column => 1, -row => 16, -columnspan => 3, -sticky => "w" );

	# server_key
	Tkx::grid( Tkx::ttk__label(".c.server_key_label", -text => "Server Key", -font => 'LabelFont'), -column => 1, -row => 17, -sticky => "w" );
	Tkx::ttk__entry(".c.server_key", -width=> 40, -textvariable=>\$server_key);
	Tkx::grid(".c.server_key", -column => 2, -row => 17, -sticky => "we");
	Tkx::grid( Tkx::ttk__label(".c.server_key_desc", -text => "The Bluestem server key.  Server keys are managed at the following URL:\nhttps://ness.uic.edu/bluestem/admin?opt=aservers"), -column => 1, -row => 18, -columnspan => 3, -sticky => "w" );

	Tkx::ttk__button(".c.install_button",
		-text => "Install Bluestem",
		-command => sub { 
			if( !$servername ) {
				Tkx::tk___messageBox(-icon => "warning", -message => "Please specify site hostname.");
			} elsif( !$server_key ) {
				Tkx::tk___messageBox(-icon => "warning", -message => "Please specify server key.");
			} else {
				do_install()
			}
		}
	);

	foreach (Tkx::SplitList(Tkx::winfo_children(".c"))) {
		Tkx::grid_configure($_, -padx => 5, -pady => 1);
	}
	Tkx::focus(".c.install_dir");

	Tkx::MainLoop();
} else {
	# Tkx interface not available
	$install_dir = ask("Target installation folder?",$install_dir);
	$install_dir =~ s{/$}{};
	$install_dir =~ s{\\}{/}g;
	$asp_install_dir ||= "$install_dir/asp";
	$cgi_install_dir ||= "$install_dir/cgi";
	$img_install_dir ||= "$install_dir/img";
	$cgi_install_dir = ask("CGI installation folder?",$cgi_install_dir);
	$cgipath = ask("CGI web path?",$cgipath);
	$img_install_dir = ask("Image installation folder?",$img_install_dir);
	$imgpath = ask("Image web path?",$imgpath);
	$asp_install_dir = ask("ASP installation folder?",$asp_install_dir);
	$asppath = ask("ASP web path?",$asppath);
	$servername = ask("Site hostname?",$servername);
	$server_key = ask("Server key?",$server_key);
	do_install();
}
exit 0
