#
#   Bluestem - University of Illinois WWW Identification Service
#
#   BlueRPC.pm  - tools for XML-RPC with shared-secret message authentication
#
#   Copyright 1996-2003 - Board of Trustees of the University of Illinois
#
#   Ed Kubaitis - ejk@uiuc.edu
#   CITES/STS/ISE
#   University of Illinois at Urbana-Champaign
#
#   Acknowledgements:
#    
#     Eric Kidd  --  XML-RPC HOWTO
#     http://xmlrpc-c.sourceforge.net/xmlrpc-howto/xmlrpc-howto.html
#
#     Ken MacLeod -- Fronter::RPC2
#     http://search.cpan.org/dist/Frontier-RPC/


package BlueRPC;
use strict;
use warnings;

use Exporter;
   our(@ISA) = qw(Exporter);
   our(@EXPORT) = qw(&bluerpc_req &bluerpc_cgi);

use Frontier::RPC2;
use Digest::HMAC_SHA1 qw(hmac_sha1_hex);
use IO::Socket;
use Net::SSLeay::Handle ();

use Env qw(
   HTTP_USER_AGENT HTTP_HOST SERVER_NAME REQUEST_METHOD CONTENT_TYPE
   CONTENT_LENGTH HTTP_COOKIE
   );


my $stripport = 1;


#----------------------------------------------------------------------
#   bluerpc_req - XML-RPC client request with message authentication
#----------------------------------------------------------------------
sub bluerpc_req {
   my($httpserver, $secret, $url, $method, @args) = @_;
   my($coder) = Frontier::RPC2->new;
   my($EOL) = "\015\012";

   #--- parse URL
   if ($url !~ m!^(https?)://([^/]+)(/.*)$!) {
      return (99901, "bluerpc_req: Invalid URL ($url)");
      }
   my($proto, $server, $path) = ($1, $2, $3);
   my($port) = $1 if $server=~ /:(\d+)$/;
   $port ||= $proto eq 'https' ? 443 : 80;


   #--- generate XML-RPC request XML and HMAC
   my($xml) = $coder->encode_call($method, @args);
   my($length) = length($xml);
   my($hmac) = hmac_sha1_hex($xml, $secret);
   
   #--- open connection to server
   my $sock;
   if( $proto eq 'http' ) {
      $sock = IO::Socket::INET->new (
                     PeerAddr => $server,
                     PeerPort => $port,
                     Proto => 'tcp'
                     );
      if (!defined($sock)) {
         return (99902, "bluerpc_req: open socket to $server:$port failed: $!")
         }
      binmode($sock);
      }
   elsif( $proto eq 'https' ) {
      tie(*SSL, "Net::SSLeay::Handle", $server, $port);
      $sock = \*SSL;
      }

   # $httpserver is used in the XML-RPC Host: header because
   # $server in $url will be localhost/127.0.0.1 when stunnel is used.
   # Strip the port info from $httpserver
   if ($stripport) {
      $httpserver =~ s/:\d+//;
   }

   #--- generate and write HTTP request
   my($headers);
   $headers = "POST $path HTTP/1.0"        . $EOL
            . "User-Agent: BlueRPC 1.0"    . $EOL
    	    . "Host: $httpserver"        . $EOL
	    . "Cookie: BlueRPC_HMAC=$hmac" . $EOL
	    . "Content-Length: $length"    . $EOL
	    . "Content-Type: text/xml"     . $EOL
	    ;
   print $sock "$headers$EOL$xml";

   #--- read and verify HTTP response status line
   my($httpstatus); # ?? declaration on same line as socket ref causes problems  
   $httpstatus = <$sock>;
   $httpstatus =~ s/\s*$// if $httpstatus; # remove trailing whitespace (ie, \r)
   if (!defined($httpstatus) || $httpstatus !~ m!^HTTP/1.\d (\d+) (.*)$!) {
      $httpstatus ||= "''";
      return (99903, "bluerpc_req: invalid HTTP status line ($httpstatus)");
      }
   my($status, $msg) = ($1, $2);
   if ($status != 200) {
      return (99904, "bluerpc_req: HTTP status ($status $msg)");
      }
   
   #--- extract values from HTTP response headers
   my ($content_type, $resplength, $resphmac) = ('', 0, '');   
   while(<$sock>) {
      chop; chop;
      last if m!^$!;
      if    (/^Content-type: (.*)$/i)                  { $content_type = $1; }
      elsif (/^Content-length: (\d+)/i)                  { $resplength = $1; }
      elsif (/^Set-cookie: BlueRPC_HMAC=([0-9a-fA-F]+)$/i) { $resphmac = $1; }
      }

   #--- verify correct MIME type for response body   
   if ($content_type ne 'text/xml') {
      close($sock);
      return (
         99905, "bluerpc_req: invalid response Content-type ($content_type)"
	 );
      }

   #--- verify Content-length header provided
   if (!$length) {
      close($sock);
      return (99906, "bluerpc_req: missing response Content-length header");
      }

   #--- verify BlueRPC HMAC provided in cookie
   if (!$resphmac) {
      close($sock);
      return ( 99907, "bluerpc_req: missing response BlueRPC HMAC cookie");
      }

   #--- read XML response body
   $xml = '';
   while(<$sock>) { $xml .= $_; }
   close($sock);

   #--- verify correct XML content-length
   $length = length($xml);   
   if ($length != $resplength) {
      return ( 
         99908,
	 "bluerpc_req: XML length error: header($resplength) read($length)"
	 );
      }

   #--- verify HMAC of XML response
   $hmac = hmac_sha1_hex($xml, $secret);
   if ($hmac ne $resphmac) {
      return ( 99909, "bluerpc_req: response HMAC error ($hmac) ($resphmac)");
      }


   #--- parse XML response and return to caller
   my($response) = $coder->decode($xml);

   if ($response->{"type"} eq 'response') {      #--- normal XML-RPC response
      return (0, $response->{'value'}[0]);
      }
   elsif ($response->{"type"} eq 'fault') {      #--- "fault" response
      return ($response->{'value'}[0]->{'faultCode'},
              $response->{'value'}[0]->{'faultString'}
	      );
      }
   else {                                         #--- anything else
      return (99910, "bluerpc_req: invalid response type ($response->{'type'}");
      } 
   }

#----------------------------------------------------------------------
#   bluerpc_cgi - XML-RPC CGI server response with message authentication
#----------------------------------------------------------------------
sub bluerpc_cgi {
   my($secret, $methods) = @_;
   my($EOL) = "\015\012";
    
   #--- perform preliminary validity checks on request headers
   defined($HTTP_USER_AGENT)   && $HTTP_USER_AGENT =~ /BlueRPC/
   && defined($HTTP_HOST)      # (may be localhost if eg stunnel used)
   && defined($REQUEST_METHOD) && $REQUEST_METHOD eq "POST"
   && defined($CONTENT_TYPE)   && $CONTENT_TYPE eq "text/xml"
   && defined($CONTENT_LENGTH) && $CONTENT_LENGTH > 0 
   || return(_http_error(400, "Bad Request"));

   #--- read request body and verify $CONTENT_LENGTH
   my $body;
   my $count = read STDIN, $body, $CONTENT_LENGTH;
   $count == $CONTENT_LENGTH || return(_http_error(400, "Bad Request")); 


   #--- verify request HMAC
   my $reqhmac = '';
   $reqhmac = $1 if $HTTP_COOKIE =~ /^BlueRPC_HMAC=([0-9a-f]+)$/i;
   $reqhmac || return(_http_error(400, "Bad Request"));
   $reqhmac eq hmac_sha1_hex($body, $secret)
      || return(_http_error(403, "Forbidden")); 

   #--- process request
   my $coder = Frontier::RPC2->new;
   my $xml = $coder->serve($body, $methods);

   #--- generate and return XML-RPC HTTP response to caller
   my $length = length($xml);
   my $hmac = hmac_sha1_hex($xml, $secret);

   my($headers);
   $headers = "Status: 200 OK"                 . $EOL
            . "Set-cookie: BlueRPC_HMAC=$hmac" . $EOL
            . "Content-type: text/xml"         . $EOL
            . "Content-length: $length"        . $EOL
            ;

   return("$headers$EOL$xml");
   }
    

#---
#    _http_error - Send HTTP error response
#---
sub _http_error {
   my ($code, $message) = @_;
   my ($headers, $html);
   my($EOL) = "\015\012";
   
   $headers = "Status: $code $message"  . $EOL
            . "Content-type: text/html" . $EOL
            ;
   ($html = <<"-30-") =~ s/^\s+//gm;
   <html>
   <head><title>$code $message</title></head>
   <body><h2>$code $message</h2></body>
   </html>
-30-
   return ("$headers$EOL$html");
   }



=head1 NAME

BlueRPC - tools for XML-RPC with shared-secret message authentication

=head1 SYNOPSIS

 use BlueRPC;

 Client:     ($rc, $result) = bluerpc_req($server, $secret, $url, $method, @args)

 CGI server: $http_response = bluerpc_cgi($secret, $methods);

=head1 DESCRIPTION

B<I<BlueRPC>> provides tools for XML-RPC client requests and
CGI server responses with messages verified by
HMAC-SHA1 message authentication codes (MACs) based on a secret
key shared between the client and server. The MACs are carried
in HTTP C<Cookie> and C<Set-Cookie> request and response headers
for a cookie named I<BlueRPC_HMAC>.

=over 4

=item I<Client:> ($rc, $result) = bluerpc_req($server, $secret, $url, $method, @args);

Make an HMAC-SHA1 authenticated XML-RPC request.

B<$rc> is a numeric return code for the request. 0 indicates
success.

B<$result> is the XML-RPC value returned by the procedure if B<$rc>
is 0 (success.) The result may be a Perl hash, array, or scalar
value.

Otherwise B<$result> contains a string describing
the error encountered. Below is a table of error codes. Codes 1-4
are returned from I<Frontier::RPC2> on the server. Codes beginning
with 999 are from I<BlueRPC> on the request host.

   1      error decoding RPC ...
   2      expected RPC methodCall, got xxxx
   3      no such method xxx
   4      error executing RPC xxx ...
   99901  bluerpc_req: Invalid URL ($url)
   99902  bluerpc_req: open socket to $server:$port failed: ...
   99903  bluerpc_req: invalid HTTP status line ($httpstatus)
   99904  bluerpc_req: HTTP status ($status $msg)
          
	  Two HTTP status codes other '200 OK' may be returned
	  by the BlueRPC CGI server: '400 Bad request' means
	  the XML-RPC request was malformed. '403 Forbidden'
	  means the HMAC computed for the request did not
	  match the HMAC in the BlueRPC_HMAC 'Cookie' HTTP
	  request header. Other HTTP status codes may be
	  returned by the HTTP server itself.

   99905  bluerpc_req: invalid response Content-type ($content_type)
   99906  bluerpc_req: missing response Content-length header
   99907  bluerpc_req: missing response BlueRPC HMAC cookie
   99908  bluerpc_req: XML length error: header($resplength) read($length)
   99909  bluerpc_req: response HMAC error ($hmac) ($resphmac)
   99910  bluerpc_req: invalid response type (xxx)   


B<$server> is the server being contacted.  This is needed
because when I<stunnel> is used, B<$url> specifies
C<localhost:port> instead of the actual server.

B<$secret> is a text string with the secret key shared between the
client and server.

B<$url> is the URL for the XML-RPC server to process the request,
Currently, only C<http> URLs are accepted. However, note that C<localhost:port>
can be used in the URL for an I<stunnel> SSL tunnel to the remote server.

B<$method> is the name of the server procedure being called
(I<methodName>), and B<@args> is a
list of arguments to pass. Arguments may be Perl hashes, arrays
or scalar values.


=back


=over 4

=item I<CGI Server:> $http_response = bluerpc_cgi($secret, $methods);

Process an HMAC-SHA1 authenticated XML-RPC request.

B<$http_response> is a text string containing a complete HTTP response
to the request, beginning with a valid CGI C<Status> header.
If the HTTP C<Status:> header is C<200 OK>, the body of the
response is C<text/xml> XML-RPC response to the request.

If the status is C<400 Bad Request>, the request
was malformed. If the status is C<403 Forbidden>, the HMAC
computed for the request body did not match the HMAC provided
in the I<BlueRPC_HMAC> C<Cookie:> header in the request. In
these two cases, the subroutine for the request is never
called. 


B<$secret> is a text string with the secret key shared between the
client and server.

B<$methods> is a hash mapping XML-RPC method names to subroutines
within this CGI that process requests for the corresponding
methods. For example:

 {'sample.echoRequest'      => \&echoRequest,
  'sample.sumAndDifference' => \&sumAndDifference}

On Windows platforms, the bluerpc_cgi caller should issue
C<binmode(STDOUT)> before sending the HTTP response.
Otherwise, extraneous CR characters may be inserted in
the XML response causing the Content-Length and/or HMAC
to be incorrect.
=back

=head1 EXAMPLE

=head2 Client

   #! /usr/local/bin/perl
   #
   use strict;
   use warnings;
   use lib ('.');
   use BlueRPC;

   my($arg1, $arg2, $toomany) = (@ARGV);
   defined($arg1) && defined($arg2) && !defined($toomany)
      && $arg1 =~ /^[-+]?\d+$/ && $arg2 =~ /^[-+]?\d+$/
      || die "usage: bluerpc_test.pl integer1 integer2\n";

   my($rc, $result) = 
      bluerpc_req(
	 "server",                                        # server being contacted
	 "secret",                                        # shared secret key
	 "http://ejk.cso.uiuc.edu/brpc/bluerpc_test.cgi", # URL
	 "test.sumAndDifference",                         # RPC name
	 $arg1, $arg2                                     # RPC args
	 );
   $rc == 0 ||  die "Error: $rc $result\n";

   print "Sum: $result->{'sum'}\nDifference: $result->{'difference'}\n";


=head2 CGI Server

   #! /usr/local/bin/perl
   #
   use strict;
   use warnings;
   use lib ('.');
   use BlueRPC;

   sub sumAndDifference {
      my($x, $y) = @_;
      return( {'sum' => $x + $y, 'difference' => $x - $y });
      }

   my ($response) = bluerpc_cgi(
      "secret",                                        # shared secret key
      {'test.sumAndDifference' => \&sumAndDifference } # RPCname-to-subroutine hash
      );

   #$response =~ /^Status: 200 OK/
   #   || &wlog("bluerpc_cgi returned error: " . (split(/\n/, $response))[0]);

   print $response;


=head1 SECURITY NOTES

The HMAC-SHA1 message authentication codes generated and
checked by B<I<BlueRPC>> provide assurances that XML-RPC
requests are made and processed only by clients and servers
with access to the shared secret key and that the request
and response bodies have not been tampered with enroute.

However, B<I<BlueRPC>> does B<I<not>> provide encryption of
requests and responses. If the data carried in requests or
responses is sensitive, an SSL I<(https)> channel should
be used for the request. If the threat model includes the
possibility of a I<man-in-the-middle> attack on the SSL
channel with DNS cache poisoning and similar tools, the client
should verify that the server's SSL certificate is
acceptable before transmitting the request.


=head1 PREREQUISITES

Frontier::RPC2, Digest::HMAC_SHA1, IO::Socket

=head1 SEE ALSO

I<XML-RPC Specification:> http://www.xmlrpc.com/spec

I<RFC 2104:> http://www.faqs.org/rfcs/rfc2104.html

I<Frontier::RPC2:> http://search.cpan.org/dist/Frontier-RPC/lib/Frontier/RPC2.pm

I<Digest::HMAC_SHA1:> http://search.cpan.org/author/GAAS/Digest-HMAC-1.01/lib/Digest/HMAC_SHA1.pm

I<XML-RPC Howto:> http://xmlrpc-c.sourceforge.net/xmlrpc-howto/xmlrpc-howto.html

I<Stunnel:> http://www.stunnel.org/

=head1 AUTHOR

Ed Kubaitis - ejk @uiuc.edu -- CITES - University of Illinois at Urbana-Champaign

=cut

1;
